﻿#Region "INFO"
'Creado por garikoitz.info en Marzo de 2025
'Entrada del Blog: https://garikoitz.info/blog/?p=2382 (Regresión No Lineal - Accord.NET vs Math.NET)
#End Region
#Region "Imports"
Imports System
Imports System.Globalization
Imports System.IO
Imports System.Reflection
'
Imports Accord
Imports Accord.Math
Imports Accord.Math.Optimization
Imports Accord.Statistics.Models.Regression
Imports Accord.Statistics.Models.Regression.Fitting

#End Region

Public Class Form1

#Region "Variables"
	Dim dt_api As DataTable
	Dim dt_api_NL As DataTable
	Dim culture As System.Globalization.CultureInfo = System.Globalization.CultureInfo.CurrentCulture
	Dim decimalSeparator As String = Globalization.CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator
	Dim T As Double() = {65.44, 73.78, 82.11, 90.45, 98.78, 107.12, 115.46, 123.79, 132.13, 140.46, 148.8, 157.13, 165.47, 173.8, 182.14}
	Dim P As Double() = {41.17, 47.32, 54.15, 61.7, 70.02, 79.15, 89.14, 100.03, 111.87, 124.7, 138.58, 153.54, 169.64, 186.9, 205.39}
	Dim A, B, C As Double
#End Region

#Region "Functions"
	Public Sub MakeGridViewDoubleBuffered(ByVal dgv As DataGridView)
		Dim dgvType As Type = dgv.[GetType]()
		Dim pis As PropertyInfo = dgvType.GetProperty("DoubleBuffered", BindingFlags.Instance Or BindingFlags.NonPublic)
		pis.SetValue(dgv, True, Nothing)
	End Sub
	Sub APIcalc()
		Try
			Dim T1, T2, T3, temp, P, Pcalc, SEC As Double
			SEC = 0.0
			For i = 0 To 14
				temp = CDbl(DGV_API_NL.Rows(i).Cells(0).Value)
				P = CDbl(DGV_API_NL.Rows(i).Cells(1).Value)
				Pcalc = 10 ^ (A - B / (temp + C))
				'T1 = 10 ^ (A - B / (temp + C))
				T2 = (+P - Pcalc)
				T3 = ((-Math.Log10(P) + (A - B / (temp + C))) ^ 2)
				DGV_API_NL.Rows(i).Cells(2).Value = Pcalc 'Pcalc
				DGV_API_NL.Rows(i).Cells(3).Value = T2 'DP psia
				DGV_API_NL.Rows(i).Cells(4).Value = T3 'Error PCT
				SEC += T3
			Next
			DGV_API_NL.Rows(15).Cells(4).Value = SEC
		Catch ex As Exception
			MsgBox(ex.ToString)
			Txt_Resul_NL.Text = ex.ToString
		End Try
	End Sub

	Public Sub CalcNoLineal_Accord()
		Try
			' --- 1. Preparar DataTable y DataGridView (DGV_API_NL) ---
			Txt_Resul_NL.Text = ""
			TXT_A.Text = ""
			TXT_B.Text = ""
			TXT_C.Text = ""

			Dim dt As Date = Now
			Txt_Resul_NL.Text = "Cálculos - " & dt.ToString("dd/MM/yyyy HH:mm:ss", Globalization.CultureInfo.InvariantCulture) & vbNewLine & vbNewLine
			Txt_Resul_NL.Text &= "Regresión No Lineal usando Levenberg-Marquardt (Accord.NET)" & vbNewLine & vbNewLine

			Dim dt_api_NL As New DataTable
			dt_api_NL.Columns.Add("Temp ºF")
			dt_api_NL.Columns.Add("P psia")
			dt_api_NL.Columns.Add("P(calc) psia")
			dt_api_NL.Columns.Add("DP psia")
			dt_api_NL.Columns.Add("Error TCP")

			' Llenar la tabla con 15 datos (índices 0 a 14)
			For i As Integer = 0 To 14
				dt_api_NL.Rows.Add(New String() {T(i).ToString(), P(i).ToString(), "0", "0", "0"})
			Next

			DGV_API_NL.DataSource = dt_api_NL
			DGV_API_NL.AutoSizeColumnsMode = DataGridViewAutoSizeColumnsMode.AllCells
			DGV_API_NL.AllowUserToOrderColumns = False
			DGV_API_NL.RowHeadersVisible = False
			DGV_API_NL.ColumnHeadersVisible = True
			Application.DoEvents()

			' --- 2. Extraer datos de la tabla en arrays ---
			Dim rowCount As Integer = dt_api_NL.Rows.Count
			Dim TData(rowCount - 1) As Double
			Dim PData(rowCount - 1) As Double

			For i As Integer = 0 To rowCount - 1
				TData(i) = CDbl(dt_api_NL.Rows(i)("Temp ºF"))
				PData(i) = CDbl(dt_api_NL.Rows(i)("P psia"))
			Next

			' --- 3. Transformar los datos: cada entrada es un vector unidimensional ---
			Dim inputs As Double()() = TData.Select(Function(t) New Double() {t}).ToArray()
			Dim output As Double() = PData

			' --- 4. Configurar la regresión no lineal ---
			' Modelo: P = 10^( A - B/(T + C) )
			Dim nls As New NonlinearLeastSquares()
			With nls
				.NumberOfParameters = 3   ' A, B, C
				' Valores iniciales (leídos de controles)
				.StartValues = {CDbl(Txt_A_ini.Text), CDbl(Txt_B_ini.Text), CDbl(Txt_C_ini.Text)}
				'
				' Función del modelo: recibe los parámetros y un vector de entrada (T)
				.Function = Function(par As Double(), inp As Double()) As Double
								Return 10 ^ (par(0) - par(1) / (inp(0) + par(2)))
							End Function
				'
				' Función del gradiente: se debe llenar el arreglo result con las derivadas parciales
				.Gradient = Sub(par As Double(), inp As Double(), result() As Double)
								Dim ln10 As Double = Math.Log(10)
								Dim pred As Double = 10 ^ (par(0) - par(1) / (inp(0) + par(2)))
								result(0) = ln10 * pred                                     ' d/dA
								result(1) = -ln10 * pred / (inp(0) + par(2))                 ' d/dB
								result(2) = (par(1) * ln10 * pred) / ((inp(0) + par(2)) ^ 2)   ' d/dC
							End Sub
			End With

			' --- 5. Configurar el algoritmo de optimización (Levenberg-Marquardt) ---
			Dim algorithm As New LevenbergMarquardt()
			With algorithm
				.MaxIterations = 1    ' Iteraciones por bloque
				.Tolerance = 0.000000000000001 ' Tolerancia para cambio en SSE -> 1e-15
			End With

			nls.Algorithm = algorithm

			' --- 6. Ejecutar el ajuste con detención temprana ---
			Dim totalMaxIter As Integer = CInt(Txt_iter.Text)
			Dim totalIter As Integer = 0
			Dim tolerance As Double = 0.000000000000001  '1e-15

			' Función local para calcular el SSE (suma de cuadrados de errores)
			Dim ComputeSSE As Func(Of Double(), Double) =
			Function(coeffs As Double()) As Double
				Dim sumSq As Double = 0
				For j As Integer = 0 To inputs.Length - 1
					Dim pred As Double = 10 ^ (coeffs(0) - coeffs(1) / (inputs(j)(0) + coeffs(2)))
					Dim resid As Double = pred - output(j)
					sumSq += resid * resid
				Next
				Return sumSq
			End Function

			Dim currentSolution() As Double = nls.StartValues
			Dim previousCost As Double = ComputeSSE(currentSolution)
			Dim currentCost As Double = previousCost

			Do
				nls.StartValues = currentSolution
				Dim regression As NonlinearRegression = nls.Learn(inputs, output)
				currentSolution = regression.Coefficients
				totalIter += 1
				currentCost = ComputeSSE(currentSolution)
				If Math.Abs(previousCost - currentCost) < tolerance Or totalIter >= 1000 Then Exit Do
				previousCost = currentCost
			Loop While totalIter < totalMaxIter

			' --- 7. Mostrar los parámetros ajustados ---
			TXT_A.Text = currentSolution(0).ToString("G5")
			TXT_B.Text = currentSolution(1).ToString("G5")
			TXT_C.Text = currentSolution(2).ToString("G5")

			A = currentSolution(0)
			B = currentSolution(1)
			C = currentSolution(2)
			APIcalc()

			' --- 8. Graficar los resultados ---
			Chart2.Series.Clear()
			Chart2.Legends.Clear()
			Chart2.Legends.Add("")
			Chart2.Titles.Clear()
			Chart2.Titles.Add("")
			Chart2.Titles(0).Font = New Font("Microsoft Sans Serif", 12, FontStyle.Bold)
			For i As Integer = 1 To 2
				Chart2.Series.Add(dt_api_NL.Columns(i).ColumnName)
				With Chart2.Series(Chart2.Series.Count - 1)
					.MarkerStyle = DataVisualization.Charting.MarkerStyle.Square
					.IsVisibleInLegend = True
					.ChartType = DataVisualization.Charting.SeriesChartType.Line
					.BorderWidth = 2
					.ToolTip = "#SERIESNAME | T: #VALX{f2} P: #VALY{f2}"
					For j As Integer = 0 To rowCount - 1
						Dim Tval As Double = CDbl(dt_api_NL.Rows(j)("Temp ºF"))
						Dim Pval As Double = CDbl(dt_api_NL.Rows(j)(i))
						.Points.AddXY(Tval, Pval)
					Next
				End With
			Next
			Chart2.ChartAreas(0).AxisX.Title = "Temperatura (ºF)"
			Chart2.ChartAreas(0).AxisY.Title = "Presión de vapor (psia)"
			Chart2.ChartAreas(0).RecalculateAxesScale()

			Txt_Resul_NL.Text &= "Total iteraciones: " & totalIter.ToString() & vbNewLine
			Txt_Resul_NL.Text &= "Costo final (SSE): " & currentCost.ToString("G5") & vbNewLine

		Catch ex As Exception
			MsgBox(ex.ToString())
		End Try
	End Sub



#End Region

#Region "Code"
	Private Sub Form1_Load(sender As Object, e As EventArgs) Handles MyBase.Load
		MakeGridViewDoubleBuffered(DGV_API_NL)
	End Sub
	Private Sub TXT_A_KeyPress(sender As Object, e As KeyPressEventArgs) Handles TXT_A.KeyPress
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub TXT_B_KeyPress(sender As Object, e As KeyPressEventArgs) Handles TXT_B.KeyPress
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub TXT_C_KeyPress(sender As Object, e As KeyPressEventArgs) Handles TXT_C.KeyPress
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Txt_Pbase_KeyPress(sender As Object, e As KeyPressEventArgs)
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Txt_SG_KeyPress(sender As Object, e As KeyPressEventArgs)
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Txt_temp_KeyPress(sender As Object, e As KeyPressEventArgs)
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Btn_calcular_Click_1(sender As Object, e As EventArgs) Handles Btn_calcular.Click
		CalcNoLineal_Accord()
	End Sub
	Private Sub DGV_API_NL_CellFormatting(sender As Object, e As DataGridViewCellFormattingEventArgs) Handles DGV_API_NL.CellFormatting
		Try
			For j = 0 To DGV_API_NL.Rows.Count - 1
				DGV_API_NL.Rows(j).Cells(1).Style.BackColor = Drawing.Color.FromArgb(65, 140, 240) '418CF0
				DGV_API_NL.Rows(j).Cells(2).Style.BackColor = Drawing.Color.FromArgb(252, 180, 65) 'FCB441
			Next
		Catch ex As Exception

		End Try
	End Sub
#End Region
End Class




