﻿#Region "INFO"
'Creado por garikoitz.info en Marzo de 2025
'Entrada del Blog: https://garikoitz.info/blog/?p=2382 (Regresión No Lineal - Accord.NET vs Math.NET)
#End Region
#Region "Imports"
Imports System.Globalization
Imports System.IO
Imports System.Reflection
Imports System.Text
Imports MathNet.Numerics
Imports MathNet.Numerics.LinearAlgebra
Imports MathNet.Numerics.Optimization
#End Region
Public Class Form1
#Region "Variables"
	Dim dt_api As DataTable
	Dim dt_api_NL As DataTable
	Dim culture As System.Globalization.CultureInfo = System.Globalization.CultureInfo.CurrentCulture
	Dim decimalSeparator As String = Globalization.CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator
	Dim T As Double() = {65.44, 73.78, 82.11, 90.45, 98.78, 107.12, 115.46, 123.79, 132.13, 140.46, 148.8, 157.13, 165.47, 173.8, 182.14}
	Dim P As Double() = {41.17, 47.32, 54.15, 61.7, 70.02, 79.15, 89.14, 100.03, 111.87, 124.7, 138.58, 153.54, 169.64, 186.9, 205.39}
	Dim A, B, C As Double
#End Region
#Region "Functions"
	Public Sub MakeGridViewDoubleBuffered(ByVal dgv As DataGridView)
		Dim dgvType As Type = dgv.[GetType]()
		Dim pis As PropertyInfo = dgvType.GetProperty("DoubleBuffered", BindingFlags.Instance Or BindingFlags.NonPublic)
		pis.SetValue(dgv, True, Nothing)
	End Sub
	Sub APIcalc()
		Try
			Dim T1, T2, T3, temp, P, Pcalc, SEC As Double
			SEC = 0.0

			For i = 0 To 14
					temp = CDbl(DGV_API_NL.Rows(i).Cells(0).Value)
					P = CDbl(DGV_API_NL.Rows(i).Cells(1).Value)
					Pcalc = 10 ^ (A - B / (temp + C))
					'T1 = 10 ^ (A - B / (temp + C))
					T2 = (+P - Pcalc)
					T3 = ((-Math.Log10(P) + (A - B / (temp + C))) ^ 2)
					DGV_API_NL.Rows(i).Cells(2).Value = Pcalc 'Pcalc
					DGV_API_NL.Rows(i).Cells(3).Value = T2 'DP psia
					DGV_API_NL.Rows(i).Cells(4).Value = T3 'Error PCT
					SEC += T3
				Next
				DGV_API_NL.Rows(15).Cells(4).Value = SEC

		Catch ex As Exception
			MsgBox(ex.ToString)
			Txt_Resul_NL.Text = ex.ToString
		End Try
	End Sub

	Public Sub CalcNoLineal_MathNet()
		Try
			' ---------------------------------------------------
			' 1. PREPARAR DataTable Y DataGridView (DGV_API_NL)
			' ---------------------------------------------------
			Txt_Resul_NL.Text = ""
			TXT_A.Text = ""
			TXT_B.Text = ""
			TXT_C.Text = ""

			Dim dt As Date = Now
			Txt_Resul_NL.Text = "Cálculos - " & dt.ToString("dd/MM/yyyy HH:mm:ss", CultureInfo.InvariantCulture) & vbNewLine & vbNewLine
			Txt_Resul_NL.Text &= "Regresión No Lineal usando Levenberg-Marquardt (Math.NET)" & vbNewLine & vbNewLine

			' Creamos la tabla para mostrar Temp, P, etc.
			Dim dt_api_NL As New DataTable
			dt_api_NL.Columns.Add("Temp ºF")
			dt_api_NL.Columns.Add("P psia")
			dt_api_NL.Columns.Add("P(calc) psia")
			dt_api_NL.Columns.Add("DP psia")
			dt_api_NL.Columns.Add("Error TCP")

			' Llenamos la tabla con los 15 datos T(i), P(i)
			For i = 0 To 14
				dt_api_NL.Rows.Add(New String() {
				T(i).ToString(),
				P(i).ToString(),
				"0", "0", "0"
			})
			Next i

			' Asignamos al DGV
			DGV_API_NL.DataSource = dt_api_NL
			DGV_API_NL.AutoSizeColumnsMode = DataGridViewAutoSizeColumnsMode.AllCells
			DGV_API_NL.AllowUserToOrderColumns = False
			DGV_API_NL.RowHeadersVisible = False
			DGV_API_NL.ColumnHeadersVisible = True
			Application.DoEvents()

			' ---------------------------------------------------
			' 2. EXTRAER DATOS DE LA TABLA => Vectores TDataVector, PDataVector
			' ---------------------------------------------------
			Dim rowCount As Integer = dt_api_NL.Rows.Count
			Dim TData As Double() = New Double(rowCount - 1) {}
			Dim PData As Double() = New Double(rowCount - 1) {}

			For i = 0 To rowCount - 1
				TData(i) = CDbl(dt_api_NL.Rows(i)("Temp ºF"))
				PData(i) = CDbl(dt_api_NL.Rows(i)("P psia"))
			Next

			' Creamos vectores Math.NET
			Dim TDataVector As Vector(Of Double) = Vector(Of Double).Build.DenseOfArray(TData)
			Dim PDataVector As Vector(Of Double) = Vector(Of Double).Build.DenseOfArray(PData)

			' ---------------------------------------------------
			' 3. DEFINIR FUNCION MODELO + JACOBIANO
			'
			' Modelo: P = 10^( A - B/(T + C) )
			' => firma: Func(Of Vector(Of Double), Double, Double)
			' => jacobiano: Func(Of Vector(Of Double), Double, Vector(Of Double))
			' ---------------------------------------------------
			Dim modelFunction As Func(Of Vector(Of Double), Double, Double) =
			Function(par, x)
				Dim A_ As Double = par(0)
				Dim B_ As Double = par(1)
				Dim C_ As Double = par(2)
				Dim Q = A_ - (B_ / (x + C_))
				Return 10.0 ^ Q
			End Function

			Dim jacobianFunction As Func(Of Vector(Of Double), Double, Vector(Of Double)) =
			Function(par, x)
				Dim A_ As Double = par(0)
				Dim B_ As Double = par(1)
				Dim C_ As Double = par(2)
				Dim ln10 As Double = Math.Log(10.0)
				Dim Q = A_ - (B_ / (x + C_))
				Dim F = 10.0 ^ Q

				Dim dFdA = ln10 * F
				Dim dFdB = -ln10 * F / (x + C_)
				Dim dFdC = ln10 * F * B_ / ((x + C_) ^ 2)

				Return Vector(Of Double).Build.DenseOfArray({dFdA, dFdB, dFdC})
			End Function

			' ---------------------------------------------------
			' 4. CREAR IObjectiveModel CON NonlinearModel
			' ---------------------------------------------------
			Dim objective As IObjectiveModel =
			ObjectiveFunction.NonlinearModel(
				modelFunction,
				jacobianFunction,
				TDataVector,
				PDataVector
			)

			' ---------------------------------------------------
			' 5. LEER ESTIMACION INICIAL Y AJUSTAR
			' ---------------------------------------------------
			Dim A_ini As Double = CDbl(Txt_A_ini.Text)
			Dim B_ini As Double = CDbl(Txt_B_ini.Text)
			Dim C_ini As Double = CDbl(Txt_C_ini.Text)
			Dim initialGuess = Vector(Of Double).Build.DenseOfArray(New Double() {A_ini, B_ini, C_ini})

			Dim lm As New LevenbergMarquardtMinimizer()
			lm.MaximumIterations = CInt(Txt_iter.Text)

			Dim result = lm.FindMinimum(objective, initialGuess)

			' ---------------------------------------------------
			' 6. Extraer parámetros ajustados => {A_Opt, B_Opt, C_Opt}
			' ---------------------------------------------------
			Dim pOptim = result.MinimizingPoint
			Dim A_Opt = pOptim(0)
			Dim B_Opt = pOptim(1)
			Dim C_Opt = pOptim(2)

			' Actualizar TextBoxes
			TXT_A.Text = A_Opt.ToString()
			TXT_B.Text = B_Opt.ToString()
			TXT_C.Text = C_Opt.ToString()

			' Mostrar info de convergencia en Txt_Resul_NL
			Txt_Resul_NL.Text &= $"Estado de convergencia: {result.ReasonForExit}" & vbNewLine
			Txt_Resul_NL.Text &= $"Iteraciones: {result.Iterations}" & vbNewLine & vbNewLine
			Txt_Resul_NL.Text &= $"Parámetros encontrados:" & vbNewLine
			Txt_Resul_NL.Text &= $"A = {A_Opt}" & vbNewLine
			Txt_Resul_NL.Text &= $"B = {B_Opt}" & vbNewLine
			Txt_Resul_NL.Text &= $"C = {C_Opt}" & vbNewLine & vbNewLine

			' ---------------------------------------------------
			' 7. CALCULAR P(calc) y actualizar la tabla
			' ---------------------------------------------------
			A = A_Opt
			B = B_Opt
			C = C_Opt
			APIcalc()

			' ---------------------------------------------------
			' 8. GRAFICAR RESULTADOS EN Chart2
			'     Se grafican las columnas "P psia" y "P(calc) psia"
			' ---------------------------------------------------
			Chart2.Series.Clear()
			Chart2.Legends.Clear()
			Chart2.Legends.Add("")
			Chart2.Titles.Clear()
			Chart2.Titles.Add("")
			Chart2.Titles(0).Font = New Font("Microsoft Sans Serif", 12, FontStyle.Bold)

			' ColIndex=1 => "P psia"
			' ColIndex=2 => "P(calc) psia"
			For colIndex As Integer = 1 To 2
				Chart2.Series.Add(dt_api_NL.Columns(colIndex).ColumnName)
				With Chart2.Series(Chart2.Series.Count - 1)
					.MarkerStyle = DataVisualization.Charting.MarkerStyle.Square
					.IsVisibleInLegend = True
					.ChartType = DataVisualization.Charting.SeriesChartType.Line
					.BorderWidth = 2
					.ToolTip = "#SERIESNAME | T: #VALX{f2} P: #VALY{f2}"

					For j = 0 To rowCount - 1
						Dim Tval = CDbl(dt_api_NL.Rows(j)("Temp ºF"))
						Dim Pval = CDbl(dt_api_NL.Rows(j)(colIndex))
						.Points.AddXY(Tval, Pval)
					Next
				End With
			Next

			' Ajustes de ejes
			Chart2.ChartAreas(0).AxisX.Title = "Temperatura (ºF)"
			Chart2.ChartAreas(0).AxisY.Title = "Presión de vapor (psia)"
			Chart2.ChartAreas(0).AxisX.LabelStyle.Format = "#"
			Chart2.ChartAreas(0).AxisY.LabelStyle.Format = "#"
			' Recalcular ejes
			Chart2.ChartAreas(0).RecalculateAxesScale()

		Catch ex As Exception
			MsgBox(ex.ToString)
		End Try
	End Sub


#End Region

#Region "Code"
	Private Sub Form1_Load(sender As Object, e As EventArgs) Handles MyBase.Load
		MakeGridViewDoubleBuffered(DGV_API_NL)
	End Sub
	Private Sub TXT_A_TextChanged_1(sender As Object, e As EventArgs) Handles TXT_A.TextChanged
		Try
			'APIcalc()
		Catch
		End Try
	End Sub
	Private Sub TXT_B_TextChanged_1(sender As Object, e As EventArgs) Handles TXT_B.TextChanged
		Try
			'APIcalc()
		Catch
		End Try
	End Sub
	Private Sub TXT_C_TextChanged_1(sender As Object, e As EventArgs) Handles TXT_C.TextChanged
		Try
			'APIcalc()
		Catch
		End Try
	End Sub
	Private Sub TXT_A_KeyPress(sender As Object, e As KeyPressEventArgs) Handles TXT_A.KeyPress
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub TXT_B_KeyPress(sender As Object, e As KeyPressEventArgs) Handles TXT_B.KeyPress
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub TXT_C_KeyPress(sender As Object, e As KeyPressEventArgs) Handles TXT_C.KeyPress
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Txt_Pbase_KeyPress(sender As Object, e As KeyPressEventArgs)
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Txt_SG_KeyPress(sender As Object, e As KeyPressEventArgs)
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Txt_temp_KeyPress(sender As Object, e As KeyPressEventArgs)
		Try
			e.Handled = Not (Char.IsDigit(e.KeyChar) Or Asc(e.KeyChar) = 8 Or ((e.KeyChar = "." Or e.KeyChar = ",") And (sender.Text.IndexOf(".") = -1 And sender.Text.IndexOf(",") = -1)))
			If e.KeyChar = "."c AndAlso decimalSeparator <> "."c Then
				e.Handled = True
				SendKeys.Send(",")
			End If
			If e.KeyChar = ","c AndAlso decimalSeparator <> ","c Then
				e.Handled = True
				SendKeys.Send(".")
			End If
		Catch ex As Exception
		End Try
	End Sub
	Private Sub Btn_calcular_Click_1(sender As Object, e As EventArgs) Handles Btn_calcular.Click
		CalcNoLineal_MathNet()
	End Sub

#End Region
End Class




